import importlib.metadata
import re
from functools import lru_cache, reduce
from importlib.metadata import Distribution

from una import alias
from una.types import ExtDeps

SUB_DEP_SEPARATORS = r"[\s!=;><\^~]"


def extract_extras(name: str) -> set[str]:
    chars = ["[", "]"]
    replacement = ","
    res = reduce(lambda acc, char: str.replace(acc, char, replacement), chars, name)
    parts = str.split(res, replacement)
    return {str.strip(p) for p in parts if p}


def extract_library_names(deps: ExtDeps) -> set[str]:
    names = {k for k in deps.items}
    with_extras = [extract_extras(n) for n in names]
    res: set[str] = set().union(*with_extras)
    return res


def known_aliases_and_sub_dependencies(deps: ExtDeps, library_alias: list[str]) -> set[str]:
    """
    Collect known aliases (packages) for third-party libraries.

    When the library origin is not from a lock-file:
    collect sub-dependencies and distribution top-namespace for each library,
    and append to the result.
    """
    lock_file = any(str.endswith(deps.source, s) for s in {".lock", ".txt"})
    third_party_libs = extract_library_names(deps)
    dists = get_distributions()
    dist_packages = distributions_packages(dists)
    custom_aliases = alias.parse(library_alias)
    sub_deps = distributions_sub_packages(dists) if not lock_file else {}
    a = alias.pick(dist_packages, third_party_libs)
    b = alias.pick(custom_aliases, third_party_libs)
    c = alias.pick(alias.KNOWN_ALIASES, third_party_libs)
    d = alias.pick(sub_deps, third_party_libs)
    e = get_packages_distributions(third_party_libs)
    return third_party_libs.union(a, b, c, d, e)


def parse_sub_package_name(dependency: str) -> str:
    parts = re.split(SUB_DEP_SEPARATORS, dependency)
    return str(parts[0])


def dist_subpackages(dist: Distribution) -> dict[str, list[str]]:
    name = dist.metadata["name"]
    dependencies = importlib.metadata.requires(name) or []
    parsed_package_names = list({parse_sub_package_name(d) for d in dependencies})
    return {name: parsed_package_names} if dependencies else {}


def map_sub_packages(acc: dict[str, list[str]], dist: Distribution) -> dict[str, list[str]]:
    return {**acc, **dist_subpackages(dist)}


def parsed_top_level_namespace(namespaces: list[str]) -> list[str]:
    return [str.replace(ns, "/", ".") for ns in namespaces]


def top_level_packages(dist: Distribution) -> list[str]:
    top_level = dist.read_text("top_level.txt")
    namespaces = str.split(top_level or "")
    return parsed_top_level_namespace(namespaces)


def mapped_packages(dist: Distribution) -> dict[str, list[str]]:
    packages = top_level_packages(dist)
    name = dist.metadata["name"]
    return {name: packages} if packages else {}


def map_packages(acc: dict[str, list[str]], dist: Distribution) -> dict[str, list[str]]:
    return {**acc, **mapped_packages(dist)}


def distributions_packages(dists: list[Distribution]) -> dict[str, list[str]]:
    """Return a mapping of top-level packages to their distributions."""
    return reduce(map_packages, dists, {})


def distributions_sub_packages(dists: list[Distribution]) -> dict[str, list[str]]:
    """Return the dependencies of each distribution."""
    init: dict[str, list[str]] = {}
    return reduce(map_sub_packages, dists, init)


@lru_cache
def get_distributions() -> list[Distribution]:
    return list(importlib.metadata.distributions())


def get_packages_distributions(project_dependencies: set[str]) -> set[str]:
    """Return the mapped top namespace from an import
    Example:
    A third-party library, such as opentelemetry-instrumentation-fastapi.
    The return value would be the mapped top namespace: opentelemetry
    Note: available for Python >= 3.10
    """
    # added in Python 3.10
    dists = importlib.metadata.packages_distributions()
    common = {k for k, v in dists.items() if project_dependencies.intersection(set(v))}
    return common.difference(project_dependencies)



from pathlib import Path

import tomlkit

from una import config, defaults
from una.types import ConfWrapper, Include, Proj, Style


def create_file(path: Path, name: str, content: str | None = None) -> Path:
    fullpath = path / name
    if content:
        with fullpath.open("w", encoding="utf-8") as f:
            f.write(content)
    else:
        fullpath.touch()
    return fullpath


def create_dir(path: Path, dir_name: str, keep: bool = False) -> Path:
    d = path / dir_name
    d.mkdir(parents=True)
    if keep:
        create_file(d, defaults.keep_file_name)
    return d


def is_int_dep_dir(p: Path) -> bool:
    return p.is_dir() and p.name not in {"__pycache__", ".venv", ".mypy_cache"}


def get_libs_dirs(root: Path, top_dir: str, ns: str) -> list[Path]:
    style = config.get_style(root)
    sub = "" if style == Style.packages else ns
    lib_dir = root / top_dir / sub
    if not lib_dir.exists():
        return []
    return [f for f in lib_dir.iterdir() if is_int_dep_dir(f)]


def get_apps_libs_names(root: Path, ns: str, top_dir: str) -> list[str]:
    dirs = get_libs_dirs(root, top_dir, ns)
    return [d.name for d in dirs]


def get_libs(root: Path, ns: str) -> list[str]:
    return get_apps_libs_names(root, ns, top_dir=defaults.libs_dir)


def get_apps(root: Path, ns: str) -> list[str]:
    return get_apps_libs_names(root, ns, defaults.apps_dir)


def collect_paths(root: Path, ns: str, int_dep: str, packages: set[str]) -> set[Path]:
    p_template = config.get_int_dep_structure(root)
    paths = {p_template.format(int_dep=int_dep, ns=ns, package=p) for p in packages}
    return {Path(root / p) for p in paths}


def collect_apps_paths(root: Path, ns: str, libs: set[str]) -> set[Path]:
    return collect_paths(root, ns, defaults.apps_dir, libs)


def collect_libs_paths(root: Path, ns: str, libs: set[str]) -> set[Path]:
    return collect_paths(root, ns, defaults.libs_dir, libs)


def update_workspace_config(path: Path, ns: str, style: Style, dependencies: str) -> None:
    pyproj = path / defaults.pyproj
    if style == Style.modules:
        with pyproj.open() as f:
            toml = tomlkit.parse(f.read())
        toml["project"]["dependencies"].append(dependencies)  # type:ignore[reportIndexIssues]
        toml["tool"]["rye"]["workspace"] = {"member": ["projects/*"]}  # type:ignore[reportIndexIssues]
        toml["tool"]["hatch"]["build"].add("dev-mode-dirs", ["libs", "apps"])  # type:ignore[reportIndexIssues]
        toml["tool"]["una"] = {"style": "modules"}  # type:ignore[reportIndexIssues]
        with pyproj.open("w", encoding="utf-8") as f:
            f.write(tomlkit.dumps(toml))  # type:ignore[reportUnknownMemberType]
    else:
        with pyproj.open() as f:
            toml = tomlkit.parse(f.read())
        toml["tool"]["rye"]["virtual"] = True  # type:ignore[reportIndexIssues]
        toml["tool"]["rye"]["workspace"] = {"member": ["apps/*", "libs/*"]}  # type:ignore[reportIndexIssues]
        toml["tool"]["una"] = {"style": "packages"}  # type:ignore[reportIndexIssues]
        with pyproj.open("w") as f:
            f.write(tomlkit.dumps(toml))  # type:ignore[reportUnknownMemberType]


def create_workspace(path: Path, ns: str, style: Style) -> None:
    app_content = defaults.app_template.format(ns=ns, lib_name=defaults.example_lib)
    lib_content = defaults.lib_template
    dependencies = f'"{defaults.example_import}"'

    update_workspace_config(path, ns, style, defaults.example_import)
    if style == Style.modules:
        create_project(path, "example_project", dependencies, defaults.example_app)

    if style == Style.packages:
        create_package(path, defaults.example_app, defaults.apps_dir, app_content, "")
        create_package(path, defaults.example_lib, defaults.libs_dir, lib_content, dependencies)
    else:
        create_module(path, defaults.example_app, defaults.apps_dir, app_content)
        create_module(path, defaults.example_lib, defaults.libs_dir, lib_content)


def parse_package_paths(packages: list[Include]) -> list[Path]:
    sorted_packages = sorted(packages, key=lambda p: p.src)
    return [Path(p.src) for p in sorted_packages]


def create_project(path: Path, name: str, dependencies: str, from_app: str) -> None:
    conf = config.load_conf(path)
    python_version = conf.project.requires_python
    ns = config.get_ns(path)

    proj_dir = create_dir(path, f"projects/{name}")
    content = defaults.projects_pyproj.format(
        ns=ns, name=name, python_version=python_version, dependencies=dependencies
    )
    create_file(
        proj_dir,
        defaults.pyproj,
        content
        + defaults.example_modules_style_project_deps.format(
            ns=ns, app_name=from_app, lib_name=defaults.example_lib
        ),
    )


def create_package(path: Path, name: str, top_dir: str, content: str, dependencies: str) -> None:
    conf = config.load_conf(path)
    python_version = conf.project.requires_python
    ns = config.get_ns(path)

    app_dir = create_dir(path, f"{top_dir}/{name}")
    ns_dir = create_dir(path, f"{top_dir}/{name}/{ns}")
    code_dir = create_dir(path, f"{top_dir}/{name}/{ns}/{name}")
    test_dir = create_dir(path, f"{top_dir}/{name}/tests")

    create_file(code_dir, "__init__.py", content)
    create_file(code_dir, "py.typed")
    is_app = top_dir == defaults.apps_dir  # TODO fix this hack
    create_file(
        test_dir, f"test_{name}_import.py", content=defaults.test_template.format(ns=ns, name=name)
    )
    pyproj_content = defaults.packages_pyproj.format(
        name=name, python_version=python_version, dependencies=dependencies
    )
    if is_app:
        create_file(ns_dir, "py.typed")  # is this necessary? basedpyright thinks so...
        if content:
            pyproj_content += defaults.example_packages_style_app_deps.format(
                ns=ns, lib_name=defaults.example_lib
            )
    create_file(
        app_dir,
        defaults.pyproj,
        pyproj_content,
    )


def create_module(path: Path, name: str, top_dir: str, content: str) -> None:
    ns = config.get_ns(path)
    code_dir = create_dir(path, f"{top_dir}/{ns}/{name}")

    create_file(code_dir, "__init__.py", content)
    create_file(code_dir, f"test_{name}.py", defaults.test_template.format(ns=ns, name=name))


def get_project_roots(root: Path) -> list[Path]:
    ws_root = config.get_workspace_root()
    style = config.get_style(ws_root)
    prefix = "projects" if style == Style.modules else "apps"
    return sorted(root.glob(f"{prefix}/*/"))


def toml_data(path: Path) -> ConfWrapper:
    return ConfWrapper(conf=config.load_conf(path), path=path)


def get_toml_files(root: Path) -> list[ConfWrapper]:
    project_files = get_project_roots(root)
    proj = [toml_data(p) for p in project_files]
    return proj


def get_projects(root: Path) -> list[Proj]:
    root_conf = config.load_conf(root)
    ns = root_conf.project.name
    confs = get_toml_files(root)
    return [
        Proj(
            name=c.conf.project.name,
            packages=config.get_project_package_includes(ns, c.conf),
            path=c.path,
            ext_deps=config.get_project_dependencies(c.conf),
        )
        for c in confs
    ]



from pathlib import Path
from typing import Annotated

from rich.console import Console
from typer import Argument, Exit, Option, Typer

from una import check, config, defaults, differ, external_deps, files, internal_deps, sync
from una.types import Options, Style

app = Typer(name="una", no_args_is_help=True, add_completion=False)
create = Typer(no_args_is_help=True)
app.add_typer(
    create,
    name="create",
    help="Commands for creating a workspace, apps, libs and projects.",
)

option_alias = Option(help="alias for third-party libraries, map install to import name")  # type: ignore[reportAny]
option_verbose = Option(help="More verbose output.")  # type: ignore[reportAny]
option_quiet = Option(help="Do not output any messages.")  # type: ignore[reportAny]


@app.command("info")
def info_command(
    alias: Annotated[str, option_alias] = "",
):
    """Info about the Una workspace."""
    root = config.get_workspace_root()
    ns = config.get_ns(root)
    options = Options(alias=str.split(alias, ",") if alias else [])

    internal_deps.int_deps_from_projects(root, ns)

    internal_deps.int_cross_deps(root, ns)

    # Deps on external libraries
    projects = internal_deps.get_projects_data(root, ns)
    filtered_projects = internal_deps.filtered_projects_data(projects)
    merged_projects_data = check.enriched_with_lock_files_data(filtered_projects, False)
    results = external_deps.external_deps_from_all(root, ns, merged_projects_data, options)
    external_deps.print_libs_in_projects(filtered_projects, options)
    if not all(results):
        raise Exit(code=1)


@app.command("diff")
def diff_command(
    since: Annotated[str, Option(help="Changed since a specific tag.")] = "",
    int_deps: Annotated[bool, Option(help="Print changed int_deps.")] = False,
):
    """Shows changed int_deps compared to the latest git tag."""
    differ.calc_diff(since, int_deps)


@app.command("sync")
def sync_command(
    check_only: Annotated[bool, Option(help="Only check, make no changes")] = False,
    quiet: Annotated[bool, option_quiet] = False,
    verbose: Annotated[bool, option_verbose] = False,
    alias: Annotated[str, option_alias] = "",
):
    """Update pyproject.toml with missing int_deps."""
    root = config.get_workspace_root()
    ns = config.get_ns(root)
    options = Options(
        quiet=quiet,
        verbose=verbose,
        alias=str.split(alias, ",") if alias else [],
    )

    if not check_only:
        projects = internal_deps.get_projects_data(root, ns)
        filtered_projects = internal_deps.filtered_projects_data(projects)
        enriched_projects = check.enriched_with_lock_files_data(filtered_projects, verbose)
        for p in filtered_projects:
            sync.sync_project_int_deps(root, ns, p, options)
            config.clear_conf_cache()

    # reload projects so any changes made by sync are picked up
    projects = internal_deps.get_projects_data(root, ns)
    filtered_projects = internal_deps.filtered_projects_data(projects)
    enriched_projects = check.enriched_with_lock_files_data(filtered_projects, verbose)
    results = {check.check_int_ext_deps(root, ns, p, options) for p in enriched_projects}
    if not all(results):
        raise Exit(code=1)


@create.command("lib")
def lib_command(
    name: Annotated[str, Argument(help="Name of the lib.")],
):
    """Creates an Una lib."""
    root = config.get_workspace_root()
    style = config.get_style(root)
    if style == Style.packages:
        files.create_package(root, defaults.example_lib, defaults.libs_dir, "", "")
    else:
        files.create_module(root, defaults.example_lib, defaults.libs_dir, "")
    console = Console(theme=defaults.una_theme)
    console.print("Success!")
    console.print(f"Created lib {name}")


@create.command("app")
def app_command(
    name: Annotated[str, Argument(help="Name of the app.")],
):
    """Creates an Una app."""
    root = config.get_workspace_root()
    style = config.get_style(root)
    if style == Style.packages:
        files.create_package(root, defaults.example_app, defaults.apps_dir, "", "")
    else:
        files.create_module(root, defaults.example_app, defaults.apps_dir, "")
    console = Console(theme=defaults.una_theme)
    console.print("Success!")
    console.print(f"Created app {name}")


@create.command("project")
def project_command(
    name: Annotated[str, Argument(help="Name of the projectlib.")],
    from_app: Annotated[str, Argument(help="Name of the app the project will use.")],
):
    """Creates an Una project."""
    root = config.get_workspace_root()
    style = config.get_style(root)
    console = Console(theme=defaults.una_theme)
    if style == Style.packages:
        console.print("You can't create projects in a Packages style workspace")
        raise Exit(code=1)

    files.create_project(root, name, "", from_app)
    console.print("Success!")
    console.print(f"Created project {name}")


@create.command("workspace")
def workspace_command(
    style: Annotated[Style, Option(help="Workspace style")] = "packages",  # type:ignore[reportArgumentType]
):
    """Creates an Una workspace in the current directory."""
    path = Path.cwd()
    root = config.get_workspace_root()
    ns = config.get_ns(root)
    files.create_workspace(path, ns, style)
    console = Console(theme=defaults.una_theme)
    console.print("Success!")
    console.print("Set up workspace in current directory.")
    console.print("Remember to delete the src/ directory")



from una import external_deps
from una.types import OrgImports


def test_calculate_diff_reports_no_diff():
    int_dep_imports = OrgImports(
        apps={"my_app": {"rich"}},
        libs={
            "one": {"rich"},
            "two": {"rich", "foo"},
            "thre": {"tomlkit"},
        },
    )
    third_party_libs = {
        "tomlkit",
        "foo",
        "requests",
        "rich",
    }
    res = external_deps.calculate_diff(int_dep_imports, third_party_libs, True)
    assert len(res) == 0


def test_calculate_diff_should_report_missing_dependency():
    expected_missing = "aws-lambda-powertools"
    int_dep_imports = OrgImports(
        apps={"my_app": {"foo"}},
        libs={
            "one": {"tomlkit"},
            "two": {"tomlkit", expected_missing, "rich"},
            "three": {"rich"},
        },
    )
    third_party_libs = {
        "tomlkit",
        "foo",
        "mypy-extensions",
        "rich",
    }
    res = external_deps.calculate_diff(int_dep_imports, third_party_libs, True)
    assert res == {expected_missing}


def test_calculate_diff_should_identify_close_match():
    int_dep_imports = OrgImports(
        apps={"my_app": {"foo"}},
        libs={
            "one": {"tomlkit"},
            "two": {"tomlkit", "aws_lambda_powertools", "rich"},
            "three": {"rich", "pyyoutube"},
        },
    )
    third_party_libs = {
        "tomlkit",
        "python-youtube",
        "foo",
        "aws-lambda-powertools",
        "rich",
    }
    res = external_deps.calculate_diff(int_dep_imports, third_party_libs, True)
    assert len(res) == 0


def test_calculate_diff_should_identify_close_match_case_insensitive():
    int_dep_imports = OrgImports(
        apps={"my_app": set()},
        libs={"one": {"PIL"}},
    )
    third_party_libs = {"pillow"}
    res = external_deps.calculate_diff(int_dep_imports, third_party_libs, True)
    assert len(res) == 0



import re
from functools import lru_cache
from pathlib import Path
from typing import cast

from una import defaults
from una.types import Conf, ExtDeps, Include, Style


def load_conf_from_str(s: str) -> Conf:
    return Conf.from_str(s)


@lru_cache
def _load_conf(path: Path) -> Conf:
    # made this private as pyright doesn't seem to like the cache decorator
    with path.open(encoding="utf-8", errors="ignore") as f:
        return load_conf_from_str(f.read())


def clear_conf_cache() -> None:
    _load_conf.cache_clear()


def load_conf(path: Path) -> Conf:
    fullpath = (path / defaults.pyproj).resolve()
    return _load_conf(fullpath)


def get_project_package_includes(namespace: str, conf: Conf) -> list[Include]:
    includes = conf.tool.una.libs
    return [Include(src=k, dst=v) for k, v in includes.items()]


def parse_pep_621_dependency(dep: str) -> dict[str, str]:
    parts = re.split(r"[\^~=!<>]", dep)
    name, *_ = parts if parts else [""]
    version = str.replace(dep, name, "")
    return {name: version} if name else {}


def get_pep_621_optional_dependencies(conf: Conf) -> list[str]:
    groups = conf.project.optional_dependencies
    matrix = [v for v in groups.values()] if isinstance(groups, dict) else []
    return sum(matrix, cast(list[str], []))


def parse_project_dependencies(conf: Conf) -> dict[str, str]:
    deps = conf.project.dependencies
    optional_deps = get_pep_621_optional_dependencies(conf)
    all_deps = deps + optional_deps
    return {k: v for dep in all_deps for k, v in parse_pep_621_dependency(dep).items()}


def get_project_dependencies(data: Conf) -> ExtDeps:
    items = parse_project_dependencies(data)
    return ExtDeps(items=items, source=defaults.pyproj)


def sanitise_name(name: str) -> str:
    return name.replace("-", "_")


def get_ns(path: Path) -> str:
    return sanitise_name(load_conf(path).project.name)


def get_style(path: Path) -> Style:
    return load_conf(path).tool.una.style


def get_tag_pattern(key: str | None) -> str:
    return "v*"


def get_tag_sort_options() -> list[str]:
    return ["-committerdate"]


def get_int_dep_structure(root: Path) -> str:
    root_conf = load_conf(root)
    style = root_conf.tool.una.style
    if style == Style.packages:
        return "{int_dep}/{package}"
    else:
        return "{int_dep}/{ns}/{package}"


def is_drive_root(cwd: Path) -> bool:
    return cwd == Path(cwd.root) or cwd == cwd.parent


def is_repo_root(cwd: Path) -> bool:
    fullpath = cwd / defaults.root_file
    return fullpath.exists()


def find_upwards(cwd: Path, name: str) -> Path | None:
    if is_drive_root(cwd):
        return None
    fullpath = cwd / name
    if fullpath.exists():
        return fullpath
    if is_repo_root(cwd):
        return None
    return find_upwards(cwd.parent, name)


def find_upwards_dir(cwd: Path, name: str) -> Path | None:
    fullpath = find_upwards(cwd, name)
    return fullpath.parent if fullpath else None


def get_workspace_root() -> Path:
    cwd = Path.cwd()
    root = find_upwards_dir(cwd, defaults.root_file)
    if not root:
        raise ValueError("Didn't find the workspace root. Expected to find a .git directory.")
    return root



from pathlib import Path

from una import defaults
from una.types import Proj


def pick_lock_file(project: Proj) -> Path | None:
    lock = Path(project.path / defaults.lock_file)
    if not lock.exists():
        return None
    return lock


def parse_name(row: str) -> str:
    parts = str.split(row, "==")
    return parts[0]


def parse_version(row: str) -> str:
    parts = str.split(row, "==")[1]
    res = str.split(parts, " ")
    return res[0]


def extract_lib_names_from_txt(path: Path) -> dict[str, str]:
    with open(path) as f:
        data = f.readlines()
    rows = (str.strip(line) for line in data)
    filtered = (row for row in rows if row and not row.startswith(("#", "-")))
    return {parse_name(row): parse_version(row) for row in filtered}


def extract_libs(path: Path) -> dict[str, str]:
    if not path.exists():
        return {}
    try:
        return extract_lib_names_from_txt(path)
    except (IndexError, KeyError, ValueError) as e:
        raise ValueError(f"Failed reading {path}: {repr(e)}") from e



from pathlib import Path

import pytest
from pytest import MonkeyPatch

from una import config
from una.types import Include

BASE_PYPROJECT = """
[project]
name = ""
version = ""
description = ""
authors = []
readme = ""
requires-python = ""
dependencies = ["fastapi~=0.109.2", "uvicorn~=0.25.0", "tomlkit"]
[tool.hatch.build]
"""


@pytest.fixture
def use_fake(monkeypatch: MonkeyPatch):
    def patch():
        monkeypatch.setattr(
            config,
            "load_conf",
            lambda _: config.load_conf_from_str(BASE_PYPROJECT),  # type: ignore[reportUnknownArgumentType]
        )

    return patch


fake_path = Path.cwd()


namespace = "unittest"
hatch_toml = f"""
{BASE_PYPROJECT}
[tool.una.libs]
"../../apps/unittest/one" = "unittest/one"
"../../libs/unittest/two" = "unittest/two"
"""

pep_621_toml_deps = f"""
{BASE_PYPROJECT}
[project.optional-dependencies]
dev = ["an-optional-lib==1.2.3", "another"]
local = ["awsglue-local-dev==1.0.0"]
"""
expected_packages = [
    Include(src="../../apps/unittest/one", dst="unittest/one"),
    Include(src="../../libs/unittest/two", dst="unittest/two"),
]


def test_get_hatch_package_includes():
    conf = config.load_conf_from_str(hatch_toml)
    res = config.get_project_package_includes(namespace, conf)
    assert res == expected_packages


def test_parse_pep_621_project_dependencies():
    expected_dependencies = {
        "fastapi": "~=0.109.2",
        "uvicorn": "~=0.25.0",
        "tomlkit": "",
        "an-optional-lib": "==1.2.3",
        "another": "",
        "awsglue-local-dev": "==1.0.0",
    }
    conf = config.load_conf_from_str(pep_621_toml_deps)
    res = config.parse_project_dependencies(conf)
    assert res == expected_dependencies



import tomllib
from pathlib import Path
from typing import Any, Literal, TypeAlias

PYPROJ = "pyproject.toml"
EXTRA_PYPROJ = Path("_extra_pyproj")
ROOT_PYPROJ_SUBDIR = Path("_root")


def load_conf(path: Path) -> dict[str, Any]:
    with (path / PYPROJ).open("rb") as fp:
        return tomllib.load(fp)


Style: TypeAlias = Literal["packages", "modules"]


def get_workspace_style(root_path: Path) -> Style:
    """
    Get the root workspace style

    Param `path` should be the path to the ap/project,
    NOT to the root workspace.
    """
    # In builds that do src -> sdist -> wheel, the root pyproject.toml file will
    # have been copied into the sdist so available for the wheel build.
    # Here we check for both in order.
    extra_root_path = EXTRA_PYPROJ / ROOT_PYPROJ_SUBDIR
    if (root_path / PYPROJ).exists():
        use_root_path = root_path
    elif (extra_root_path / PYPROJ).exists():
        use_root_path = extra_root_path
    else:
        raise ValueError("No root pyproject to determine workspace style")
    root_conf = load_conf(use_root_path)
    try:
        style: Style = root_conf["tool"]["una"]["style"]
        return style
    except KeyError as e:
        raise KeyError(
            "Root workspace pyproject.toml needs '[tool.una]' with style specified"
        ) from e



import importlib.metadata

from una import distributions
from una.types import ExtDeps


class FakeDist:
    def __init__(self, name: str, data: str):
        self.data = data
        self.metadata = {"name": name}

    def read_text(self, *args: object) -> str:
        return self.data


def test_distribution_packages_parse_contents_of_top_level_txt():
    dists = [FakeDist("python-jose", "jose\njose/backends\n")]
    res = distributions.distributions_packages(dists)  # type: ignore[reportArgumentType]
    expected_dist = "python-jose"
    expected_packages = ["jose", "jose.backends"]
    assert res.get(expected_dist) is not None
    assert res[expected_dist] == expected_packages


def test_parse_package_name_from_dist_requires():
    expected = {
        "greenlet": "greenlet !=0.4.17",
        "mysqlclient": "mysqlclient >=1.4.0 ; extra == 'mysql'",
        "typing-extensions": "typing-extensions>=4.6.0",
        "pymysql": "pymysql ; extra == 'pymysql'",
        "one": "one<=0.4.17",
        "two": "two^=0.4.17",
        "three": "three~=0.4.17",
    }
    for k, v in expected.items():
        assert k == distributions.parse_sub_package_name(v)


def test_distribution_sub_packages():
    dists = list(importlib.metadata.distributions())
    res = distributions.distributions_sub_packages(dists)
    expected_dist = "typer"
    expected_sub_package = "typing-extensions"
    assert res.get(expected_dist) is not None
    assert expected_sub_package in res[expected_dist]


def test_parse_third_party_library_name():
    fake_project_deps = ExtDeps(
        items={
            "python": "^3.10",
            "fastapi": "^0.110.0",
            "uvicorn[standard]": "^0.27.1",
            "python-jose[cryptography]": "^3.3.0",
            "hello[world, something]": "^3.3.0",
        },
        source="pyproject.toml",
    )
    expected = {
        "python",
        "fastapi",
        "uvicorn",
        "standard",
        "python-jose",
        "cryptography",
        "hello",
        "world",
        "something",
    }
    res = distributions.extract_library_names(fake_project_deps)
    assert res == expected



from rich.theme import Theme

una_theme = Theme(
    {
        "data": "#999966",
        "proj": "#8A2BE2",
        "lib": "#32CD32",
        "app": "#6495ED",
    }
)

keep_file_name = ".keep"
lock_file = "requirements.lock"
root_file = ".git"
pyproj = "pyproject.toml"

libs_dir = "libs"
apps_dir = "apps"
proj_dir = "projects"

example_app = "printer"
example_lib = "greeter"

example_import = "cowsay-python==1.0.2"

packages_pyproj = """\
[project]
name = "{name}"
version = "0.1.0"
description = ""
authors = []
dependencies = [{dependencies}]
requires-python = "{python_version}"
dynamic = ["una"]  # needed for hatch-una metadata hook to work

[build-system]
requires = ["hatchling", "hatch-una"]
build-backend = "hatchling.build"

[tool.hatch.metadata]
allow-direct-references = true

[tool.rye]
managed = true
dev-dependencies = []

[tool.hatch.build.hooks.una-build]
[tool.hatch.metadata.hooks.una-meta]

[tool.una.libs]
"""

projects_pyproj = """\
[project]
name = "{name}"
version = "0.1.0"
description = ""
authors = []
dependencies = [{dependencies}]
requires-python = "{python_version}"

[build-system]
requires = ["hatchling", "hatch-una"]
build-backend = "hatchling.build"

[tool.hatch.metadata]
allow-direct-references = true

[tool.hatch.build]
packages = ["{ns}"]

[tool.hatch.build.hooks.una-build]

[tool.una.libs]
"""

example_packages_style_app_deps = """\
"../../libs/{lib_name}/{ns}/{lib_name}" = "{ns}/{lib_name}"
"""

example_modules_style_project_deps = """\
"../../apps/{ns}/{app_name}" = "{ns}/{app_name}"
"../../libs/{ns}/{lib_name}" = "{ns}/{lib_name}"
"""

app_template = """\
import {ns}.{lib_name} as {lib_name}


def run() -> None:
    print({lib_name}.greet())
"""

lib_template = """\
import cowsay


def greet() -> str:
    return cowsay.say("Hello from una!")
"""

test_template = """\
from {ns} import {name}


def test_import():
    assert {name}
"""



from una import stdlib


def test_stdlib_3_11():
    assert stdlib.py310.difference(stdlib.py311) == {"binhex"}
    assert stdlib.py311.difference(stdlib.py310) == {
        "tomllib",
        "_tkinter",
        "sitecustomize",
        "usercustomize",
    }


def test_stdlib_3_12():
    assert stdlib.py311.difference(stdlib.py312) == {
        "asynchat",
        "asyncore",
        "distutils",
        "imp",
        "smtpd",
    }
    assert stdlib.py312.difference(stdlib.py311) == set()



from functools import reduce
from typing import cast


def _to_key_with_values(acc: dict[str, list[str]], alias: str) -> dict[str, list[str]]:
    k, v = str.split(alias, "=")
    values = [str.strip(val) for val in str.split(v, ",")]
    return {**acc, **{k: values}}


def parse(aliases: list[str]) -> dict[str, list[str]]:
    """Parse a list of aliases defined as key=value(s) into a dictionary"""
    return reduce(_to_key_with_values, aliases, {})


def pick(aliases: dict[str, list[str]], keys: set[str]) -> set[str]:
    matrix = [v for k, v in aliases.items() if k in keys]
    flattened = sum(matrix, cast(list[str], []))
    return set(flattened)


KNOWN_ALIASES = {
    "beautifulsoup4": ["bs4"],
    "pillow": ["PIL"],
    "scikit-learn": ["sklearn"],
    "scikit-image": ["skimage"],
    "opencv-python": ["cv2"],
    "python-ffmpeg": ["ffmpeg"],
    "pycryptodome": ["Crypto"],
    "pycryptodomex": ["Cryptodome"],
    "pyserial": ["serial"],
    "python-multipart": ["multipart"],
    "pyusb": ["usb"],
    "pyyaml": ["yaml"],
}



from functools import reduce  # noqa: I001
from pathlib import Path

from rich import box
from rich.console import Console
from rich.padding import Padding
from rich.table import Table

from una import check, config, files, defaults
from una.types import Imports, Include, IntDeps, OrgImports, Proj, Style


def int_cross_deps(root: Path, ns: str) -> None:
    apps = set(files.get_apps(root, ns))
    libs = set(files.get_libs(root, ns))
    int_dep_imports = get_int_dep_imports(root, ns, apps, libs)
    imports = {**int_dep_imports.apps, **int_dep_imports.libs}

    flattened = flatten_imports(imports)
    imported_apps = sorted({b for b in flattened if b in apps})
    imported_libs = sorted({c for c in flattened if c in libs})
    imported_int_deps = imported_libs + imported_apps
    table = Table(box=box.SIMPLE_HEAD)
    table.add_column("[data]int_dep[/]")
    cols = create_columns(imported_apps, imported_libs)
    rows = create_rows(apps, libs, imports, imported_int_deps)
    for col in cols:
        table.add_column(col, justify="center")
    for row in rows:
        table.add_row(*row)
    console = Console(theme=defaults.una_theme)
    console.print(Padding("[data]Internal libs in libs and apps[/]", (1, 0, 0, 0)))
    console.print(table, overflow="ellipsis")


def get_int_dep_imports(root: Path, ns: str, apps: set[str], libs: set[str]) -> OrgImports:
    apps_paths = files.collect_apps_paths(root, ns, apps)
    comp_paths = files.collect_libs_paths(root, ns, libs)
    int_dep_imports_in_apps = check.extract_int_deps(apps_paths, ns)
    int_dep_imports_in_libs = check.extract_int_deps(comp_paths, ns)
    return OrgImports(
        apps=check.with_unknown_libs(root, ns, int_dep_imports_in_apps),
        libs=check.with_unknown_libs(root, ns, int_dep_imports_in_libs),
    )


def to_col(int_dep: str, tag: str) -> str:
    return f"[{tag}]{int_dep}[/]"


def int_dep_status_matrix(int_deps: set[str], int_dep_name: str, imported: str) -> str:
    status = ":heavy_check_mark:" if imported in int_deps and imported != int_dep_name else "-"
    return f"[data]{status}[/]"


def to_row(name: str, tag: str, int_dep_imports: Imports, imported: list[str]) -> list[str]:
    int_deps = int_dep_imports.get(name) or set()
    statuses = [int_dep_status_matrix(int_deps, name, i) for i in imported]
    return [f"[{tag}]{name}[/]"] + statuses


def flatten_import(acc: set[str], kv: tuple[str, set[str]]) -> set[str]:
    key = kv[0]
    values = kv[1]
    res: set[str] = set().union(acc, values.difference({key}))
    return res


def flatten_imports(int_dep_imports: Imports) -> set[str]:
    """
    Flatten the dict into a set of imports.

    The actual int_dep filtered away when existing as an import.
    """
    return reduce(flatten_import, int_dep_imports.items(), set())


def create_columns(imported_apps: list[str], imported_libs: list[str]) -> list[str]:
    app_cols = [to_col(int_dep, "app") for int_dep in imported_apps]
    lib_cols = [to_col(int_dep, "lib") for int_dep in imported_libs]
    return lib_cols + app_cols


def create_rows(
    apps: set[str], libs: set[str], import_data: Imports, imported: list[str]
) -> list[list[str]]:
    app_rows = [to_row(b, "app", import_data, imported) for b in sorted(apps)]
    lib_rows = [to_row(c, "lib", import_data, imported) for c in sorted(libs)]
    return lib_rows + app_rows


def int_deps_from_projects(root: Path, ns: str) -> None:
    apps = files.get_apps(root, ns)
    libs = files.get_libs(root, ns)
    if not libs and not apps:
        return
    projects = get_int_deps_in_projects(root, libs, apps, ns)
    table = build_int_deps_in_projects_table(projects, apps, libs)
    console = Console(theme=defaults.una_theme)
    console.print(Padding("[data]Internal deps in projects[/]", (1, 0, 0, 0)))
    console.print(table, overflow="ellipsis")


def get_matching_int_deps(paths: list[Path], int_deps: list[str], namespace: str) -> list[str]:
    paths_in_namespace = [p.name for p in paths if p.parent.name == namespace]
    res = set(int_deps).intersection(paths_in_namespace)
    return sorted(list(res))


def get_project_int_deps(
    project_packages: list[Include],
    libs_paths: list[str],
    apps_paths: list[str],
    namespace: str,
    self_name: str,
    add_self: bool,
) -> IntDeps:
    paths = files.parse_package_paths(project_packages)
    libs_in_project = get_matching_int_deps(paths, libs_paths, namespace)
    apps_in_project = get_matching_int_deps(paths, apps_paths, namespace)
    if add_self:
        apps_in_project.append(self_name)
    return IntDeps(libs=libs_in_project, apps=apps_in_project)


def get_int_deps_in_projects(
    root: Path, libs_paths: list[str], apps_paths: list[str], namespace: str
) -> list[Proj]:
    packages = files.get_projects(root)
    ws_root = config.get_workspace_root()
    style = config.get_style(ws_root)
    add_self = style == Style.packages
    res = [
        Proj(
            name=p.name,
            packages=p.packages,
            path=p.path,
            ext_deps=p.ext_deps,
            int_deps=get_project_int_deps(
                p.packages,
                libs_paths,
                apps_paths,
                namespace,
                p.name,
                add_self,
            ),
        )
        for p in packages
    ]
    return res


def get_projects_data(root: Path, ns: str) -> list[Proj]:
    app_names = files.get_apps(root, ns)
    lib_names = files.get_libs(root, ns)
    return get_int_deps_in_projects(root, lib_names, app_names, ns)


def filtered_projects_data(projects: list[Proj]) -> list[Proj]:
    return [p for p in projects if Path.cwd().name in p.path.as_posix()]


def int_dep_status_projects(int_dep: str, int_deps: list[str], for_info: bool) -> str:
    emoji = ":heavy_check_mark:" if for_info else ":gear:"
    status = emoji if int_dep in int_deps else "-"
    return f"[data]{status}[/]"


def build_int_deps_in_projects_table(
    projects_data: list[Proj],
    apps: list[str],
    libs: list[str],
    for_info: bool = True,
) -> Table:
    table = Table(box=box.SIMPLE_HEAD)
    table.add_column("[data]int_dep[/]")
    proj_cols = [f"[proj]{p.name}[/]" for p in projects_data]
    for col in proj_cols:
        table.add_column(col, justify="center")
    for int_dep in sorted(libs):
        statuses = [
            int_dep_status_projects(int_dep, p.int_deps.libs, for_info) for p in projects_data
        ]
        cols = [f"[lib]{int_dep}[/]"] + statuses
        table.add_row(*cols)
    for int_dep in sorted(apps):
        statuses = [
            int_dep_status_projects(int_dep, p.int_deps.apps, for_info) for p in projects_data
        ]
        cols = [f"[app]{int_dep}[/]"] + statuses
        table.add_row(*cols)
    return table



from pathlib import Path

from pdm.backend.hooks import Context

from pdm_una import util


def force_include(context: Context) -> None:
    """
    Force-include all needed internal monorepo dependencies.
    """
    print("una: Injecting internal dependencies")
    context.ensure_build_dir()

    build_dir = context.build_dir

    # load the config for this app/project
    path = Path(context.root)
    conf = util.load_conf(path)
    name: str = conf["project"]["name"]

    try:
        int_deps: dict[str, str] = conf["tool"]["una"]["libs"]
    except KeyError as e:
        raise KeyError(
            f"App/project '{name}' is missing '[tool.una.libs]' in pyproject.toml"
        ) from e

    # need to determine workspace style (packages or modules)
    # as packages style needs dependencies' pyproject.tomls to be included
    # so that they're available in src -> sdist -> wheel builds
    via_sdist = (util.EXTRA_PYPROJ / util.ROOT_PYPROJ_SUBDIR / util.PYPROJ).exists()
    if via_sdist:
        # nothing to do as everything should already be included in sdist...
        return

    root_path = path.parents[1]
    style = util.get_workspace_style(root_path)

    if not int_deps:
        if style == "packages":
            # this is fine, the app doesn't import anything internally
            return
        else:
            # this is an empty project, useless and accidental
            raise ValueError(f"Project '{name}' has no dependencies")

    # make sure all int_deps exist
    found = [Path(k) for k in int_deps if (path / k).exists()]
    missing = set(int_deps) - set(str(p) for p in found)
    if len(missing) > 0:
        missing_str = ", ".join(missing)
        raise ValueError(f"Could not find these paths: {missing_str}")

    # need to add the root workspace pyproject.toml so that in src -> sdist -> wheel builds,
    # we can still determine the style (for packages style)
    util.copy_file(
        root_path / util.PYPROJ,
        build_dir / util.EXTRA_PYPROJ / util.ROOT_PYPROJ_SUBDIR / util.PYPROJ,
    )
    for src_str, dst_str in int_deps.items():
        src = Path(src_str)
        destination = build_dir / dst_str
        util.copy_tree(src, destination)
        if style == "packages":
            # need these so src->sdist->wheel builds can access them for external deps
            util.copy_file(
                src.parents[1] / util.PYPROJ,
                build_dir / util.EXTRA_PYPROJ / src.name / util.PYPROJ,
            )



from pathlib import Path
from typing import Any

from hatchling.builders.config import BuilderConfig
from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from hatchling.plugin import hookimpl

from hatch_una import util


class UnaBuildHook(BuildHookInterface[BuilderConfig]):
    """
    Force-include all needed internal monorepo dependencies.
    """

    PLUGIN_NAME = "una-build"

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        print("una: Injecting internal dependencies")

        # load the config for this app/project
        path = Path(self.root)
        conf = util.load_conf(path)
        name: str = conf["project"]["name"]

        try:
            int_deps: dict[str, str] = conf["tool"]["una"]["libs"]
        except KeyError as e:
            raise KeyError(
                f"App/project '{name}' is missing '[tool.una.libs]' in pyproject.toml"
            ) from e

        # need to determine workspace style (packages or modules)
        # as packages style needs dependencies' pyproject.tomls to be included
        # so that they're available in src -> sdist -> wheel builds
        root_path = path.parents[1]
        style = util.get_workspace_style(root_path)

        if not int_deps:
            if style == "packages":
                # this is fine, the app doesn't import anything internally
                return
            else:
                # this is an empty project, useless and accidental
                raise ValueError(f"Project '{name}' has no dependencies")

        # make sure all int_deps exist
        found = [Path(k) for k in int_deps if (path / k).exists()]
        missing = set(int_deps) - set(str(p) for p in found)
        if len(missing) > 0:
            missing_str = ", ".join(missing)
            raise ValueError(f"Could not find these paths: {missing_str}")

        # need to add the root workspace pyproject.toml so that in src -> sdist -> wheel builds,
        # we can still determine the style (for packages style)
        add_root_pyproj = {
            str(root_path / util.PYPROJ): str(
                util.EXTRA_PYPROJ / util.ROOT_PYPROJ_SUBDIR / util.PYPROJ
            )
        }
        if style == "packages":
            add_packages_pyproj = {
                str(f.parents[1] / util.PYPROJ): str(util.EXTRA_PYPROJ / f.name / util.PYPROJ)
                for f in found
            }
        else:
            add_packages_pyproj = {}

        build_data["force_include"] = {
            **build_data["force_include"],
            **int_deps,
            **add_root_pyproj,
            **add_packages_pyproj,
        }


@hookimpl
def hatch_register_build_hook() -> type[UnaBuildHook]:
    return UnaBuildHook



from copy import deepcopy
from pathlib import Path

from rich.console import Console

from una import config, defaults, distributions, external_deps, files, lock_files, parse
from una.types import CheckReport, ExtDeps, Imports, Options, OrgImports, Proj, Style


def check_int_ext_deps(root: Path, ns: str, project: Proj, options: Options) -> bool:
    name = config.sanitise_name(project.name)
    int_dep_imports, ext_dep_imports = collect_all_imports(root, ns, project)
    collected_libs = collect_known_aliases(project, options)
    details = create_report(
        project,
        int_dep_imports,
        ext_dep_imports,
        collected_libs,
    )
    num_apps = len(project.int_deps.apps)
    res = all([num_apps == 1, not details.int_dep_diff, not details.ext_dep_diff])
    if not options.quiet:
        print_one_app(num_apps, name)
        print_missing_deps(details.int_dep_diff, name)
        print_missing_deps(details.ext_dep_diff, name)
        if options.verbose:
            print_int_dep_imports(details.int_dep_imports)
            print_int_dep_imports(details.ext_dep_imports)
    return res


def with_ext_deps_from_lock_file(project: Proj) -> Proj:
    lock_file_path = lock_files.pick_lock_file(project)
    if not lock_file_path:
        return project
    ext_deps = lock_files.extract_libs(lock_file_path)
    project = deepcopy(project)
    project.ext_deps = ExtDeps(source=str(lock_file_path), items=ext_deps)
    return project


def enriched_with_lock_file_data(project: Proj, is_verbose: bool) -> Proj:
    try:
        return with_ext_deps_from_lock_file(project)
    except ValueError as e:
        if is_verbose:
            print(f"{project.name}: {e}")
        return project


def enriched_with_lock_files_data(projects: list[Proj], is_verbose: bool) -> list[Proj]:
    return [enriched_with_lock_file_data(p, is_verbose) for p in projects]


def collect_known_aliases(project: Proj, options: Options) -> set[str]:
    return distributions.known_aliases_and_sub_dependencies(project.ext_deps, options.alias)


def only_int_dep_imports(imports: set[str], top_ns: str) -> set[str]:
    return {i for i in imports if i.startswith(top_ns)}


def only_int_dep_name(int_dep_imports: set[str]) -> set[str]:
    res = [i.split(".") for i in int_dep_imports]
    return {i[1] for i in res if len(i) > 1}


def extract_int_dep_imports(all_imports: Imports, top_ns: str) -> Imports:
    only_int = {k: only_int_dep_imports(v, top_ns) for k, v in all_imports.items()}
    return {k: only_int_dep_name(v) for k, v in only_int.items() if v}


def extract_int_deps(paths: set[Path], ns: str) -> Imports:
    all_imports = parse.fetch_all_imports(paths)
    return extract_int_dep_imports(all_imports, ns)


def with_unknown_libs(root: Path, ns: str, int_dep_imports: Imports) -> Imports:
    keys = set(int_dep_imports.keys())
    values: set[str] = set().union(*int_dep_imports.values())
    unknowns = values.difference(keys)
    if not unknowns:
        return int_dep_imports
    paths = files.collect_libs_paths(root, ns, unknowns)
    extracted = extract_int_deps(paths, ns)
    if not extracted:
        return int_dep_imports
    collected = {**int_dep_imports, **extracted}
    return with_unknown_libs(root, ns, collected)


def diff(known_int_deps: set[str], apps: set[str], libs: set[str]) -> set[str]:
    int_deps: set[str] = set().union(apps, libs)
    return known_int_deps.difference(int_deps)


def imports_diff(int_dep_imports: OrgImports, apps: set[str], libs: set[str]) -> set[str]:
    flattened_apps: set[str] = set().union(*int_dep_imports.apps.values())
    flattened_libs: set[str] = set().union(*int_dep_imports.libs.values())
    flattened_imports: set[str] = set().union(flattened_apps, flattened_libs)
    return diff(flattened_imports, apps, libs)


def fetch_int_dep_imports(root: Path, ns: str, all_imports: Imports) -> Imports:
    extracted = extract_int_dep_imports(all_imports, ns)
    res = with_unknown_libs(root, ns, extracted)
    return res


def print_int_dep_imports(int_dep_imports: OrgImports) -> None:
    console = Console(theme=defaults.una_theme)
    apps_imports = int_dep_imports.apps
    libs_imports = int_dep_imports.libs
    int_deps = {**apps_imports, **libs_imports}
    for key, values in int_deps.items():
        imports_in_int_dep = values.difference({key})
        if not imports_in_int_dep:
            continue
        joined = ", ".join(imports_in_int_dep)
        message = f":information: [data]{key}[/] is importing [data]{joined}[/]"
        console.print(message)


def print_one_app(num_apps: int, project_name: str) -> None:
    if num_apps == 1:
        return
    console = Console(theme=defaults.una_theme)
    console.print(f"Projects must include exactly ONE app, but {project_name} has {num_apps}")


def print_missing_deps(diff: set[str], project_name: str) -> None:
    if not diff:
        return
    console = Console(theme=defaults.una_theme)
    missing = ", ".join(sorted(diff))
    console.print(f":thinking_face: Cannot locate {missing} in {project_name}")


def collect_all_imports(root: Path, ns: str, project: Proj) -> tuple[OrgImports, OrgImports]:
    app_pkgs = {b for b in project.int_deps.apps}
    libs_pkgs = {c for c in project.int_deps.libs}
    apps_paths = files.collect_apps_paths(root, ns, app_pkgs)
    libs_paths = files.collect_libs_paths(root, ns, libs_pkgs)
    all_imports_in_apps = parse.fetch_all_imports(apps_paths)
    all_imports_in_libs = parse.fetch_all_imports(libs_paths)
    int_dep_imports = OrgImports(
        apps=fetch_int_dep_imports(root, ns, all_imports_in_apps),
        libs=fetch_int_dep_imports(root, ns, all_imports_in_libs),
    )
    ext_dep_imports = OrgImports(
        apps=external_deps.extract_ext_dep_imports(all_imports_in_apps, ns),
        libs=external_deps.extract_ext_dep_imports(all_imports_in_libs, ns),
    )
    return int_dep_imports, ext_dep_imports


def create_report(
    project: Proj,
    int_dep_imports: OrgImports,
    ext_dep_imports: OrgImports,
    third_party_libs: set[str],
) -> CheckReport:
    app_pkgs = {b for b in project.int_deps.apps}
    lib_pkgs = {c for c in project.int_deps.libs}
    int_dep_diff = imports_diff(int_dep_imports, app_pkgs, lib_pkgs)

    root = config.get_workspace_root()
    style = config.get_style(root)
    include_libs = style == Style.modules
    ext_dep_diff = external_deps.calculate_diff(ext_dep_imports, third_party_libs, include_libs)

    return CheckReport(
        int_dep_imports=int_dep_imports,
        ext_dep_imports=ext_dep_imports,
        int_dep_diff=int_dep_diff,
        ext_dep_diff=ext_dep_diff,
    )



from pathlib import Path

from una import check
from una.types import ExtDeps, Options, Proj


def test_collect_known_aliases_and_sub_dependencies():
    fake_project_data = Proj(
        name="",
        packages=[],
        path=Path(),
        ext_deps=ExtDeps(
            items={"typer": "1", "hello-world-library": "2"},
            source="unit-test",
        ),
    )
    fake_options = Options(alias=["hello-world-library=hello"])
    res = check.collect_known_aliases(fake_project_data, fake_options)
    assert "typer" in res
    assert "typing-extensions" in res
    assert "hello" in res
